""" Module for web interface
This mini web interfaces is based on Flask.
Theme help: https://bootswatch.com/darkly/
"""
from flask import Flask, render_template # pylint: disable=C0301
from waitress import serve
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.web.controller.customizing import CustomizingEditController,\
    CustomizingListController, CustomizingSaveController
from databus.web.controller.export import \
    ExportController, ExportGetDictController, ExportExeController
from databus.web.controller.home import HomeController, AboutController
from databus.web.controller.log import LogDisplayController, LogListController, LogPurgeController
from databus.web.controller.login import LoginAttemptController, LogoffController
from databus.web.controller.passenger import PassengerExpediteController, PassengerListController
from databus.web.controller.peek import PeekAttachmentController, PeekController
from databus.web.controller.queue import QueueAttachmentController, QueueDisplayController,\
    QueueListController, QueuePurgeController, QueueStatusUpdateController
from databus.web.controller.user import UserListController, UserTokenRevokeController
from databus.web.controller.shutdown import ShutdownController, ShutdownExeController
from databus.web import util

##############################
# Main stuff
##############################

_APP = Flask(__name__)
_APP.secret_key = "databus"
_APP.config["CACHE_TYPE"] = "null"
_APP.config["JSONIFY_PRETTYPRINT_REGULAR"] = True
_DISPATCHER: AbstractDispatcher


def run_web_server(dispatcher: AbstractDispatcher):
    """ Starts the Flask Web Server """
    global _DISPATCHER # pylint: disable=W0603
    _DISPATCHER = dispatcher
    serve(_APP, port=dispatcher.ticket.web_server_port)
    #_APP.run(port=dispatcher.ticket.web_server_port)

@_APP.context_processor
def databus_context_processor():
    """ Returns if the user is root or not, used in default template """
    return util.get_authorization()

##############################
# Home page
##############################

@_APP.route("/")
def _home():
    return HomeController(_DISPATCHER).execute()

@_APP.route("/about")
def _about():
    return AboutController(_DISPATCHER).execute()

##############################
# Log pages
##############################

@_APP.route("/log_list")
def _log_list():
    return LogListController(_DISPATCHER).execute()

@_APP.route("/log_display")
def _log_display():
    return LogDisplayController(_DISPATCHER).execute()

@_APP.route("/log_purge")
def _log_purge():
    return LogPurgeController(_DISPATCHER).execute()

##############################
# Queue pages
##############################

@_APP.route("/queue_list")
def _queue_list():
    return QueueListController(_DISPATCHER).execute()

@_APP.route("/queue_display")
def _queue_display():
    return QueueDisplayController(_DISPATCHER).execute()

@_APP.route("/queue_attachment")
def _queue_attachment():
    return QueueAttachmentController(_DISPATCHER).execute()

@_APP.route("/queue_status_update")
def _queue_status_update(): # pylint: disable=R0912
    return QueueStatusUpdateController(_DISPATCHER).execute()

@_APP.route("/queue_purge")
def _queue_purge():
    return QueuePurgeController(_DISPATCHER).execute()

##############################
# Passengers
##############################

@_APP.route("/passenger_list")
def _passenger_list():
    return PassengerListController(_DISPATCHER).execute()

@_APP.route("/passenger_expedite")
def _passenger_expedite():
    return PassengerExpediteController(_DISPATCHER).execute()

##############################
# Peek
##############################

@_APP.route("/peek")
def _peek():
    return PeekController(_DISPATCHER).execute()

@_APP.route("/peek_attachment")
def _peek_attachment():
    return PeekAttachmentController(_DISPATCHER).execute()

##############################
# Login
##############################

@_APP.route("/login")
def _login():
    return render_template("login.html")

@_APP.route("/login_attempt", methods=["POST"])
def _login_attempt():
    return LoginAttemptController(_DISPATCHER).execute()

@_APP.route("/logoff")
def _logoff():
    return LogoffController(_DISPATCHER).execute()

##############################
# Users
##############################

@_APP.route("/user_list")
def _user_list():
    return UserListController(_DISPATCHER).execute()

@_APP.route("/user_token_revoke")
def _user_token_revoke():
    return UserTokenRevokeController(_DISPATCHER).execute()

##############################
# Customizing
##############################

@_APP.route("/customizing_list")
def _customizing_list():
    return CustomizingListController(_DISPATCHER).execute()

@_APP.route("/customizing_edit")
def _customizing_edit():
    return CustomizingEditController(_DISPATCHER).execute()

@_APP.route("/customizing_save", methods=["POST"])
def _customizing_save():
    return CustomizingSaveController(_DISPATCHER).execute()

##############################
# Export
##############################

@_APP.route("/export")
def _export():
    return ExportController(_DISPATCHER).execute()

@_APP.route("/export_get_dict")
def _export_get_dict():
    return ExportGetDictController(_DISPATCHER).execute()

@_APP.route("/export_exe", methods=["POST"])
def _export_exe():
    return ExportExeController(_DISPATCHER).execute()

##############################
# Shutdown
##############################

@_APP.route("/shutdown")
def _shutdown():
    return ShutdownController(_DISPATCHER).execute()

@_APP.route("/shutdown_exe")
def _shutdown_exe():
    return ShutdownExeController(_DISPATCHER).execute()



""" Json database implementation module for client """
from copy import copy
import json
from os import path, scandir
import pathlib
from shutil import rmtree
from typing import List
from databus.client.client import Client, ClientError, ClientPassenger
from databus.client.user import User, Credential, str_to_role, Role
from databus.database.difference_check import Action, DifferenceChecker, TableKey
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.json_db.json_path_builder import JsonPathBuilder

class JsonClient:
    """ JSON database implementation for client """
    def __init__(self, args: JsonDatabaseArguments):
        self._args = args

    @property
    def all_clients(self) -> List[Client]:
        """ Returns all clients """
        output = []
        for client_directory in self.client_directories:
            config_json = self.get_config_as_json(client_directory)
            client_passengers = []
            for passenger_json in config_json["passengers"]:
                client_passenger = ClientPassenger(
                    p_name=passenger_json["name"],
                    p_puller_modules=passenger_json["pullers"],
                    p_queue_module=passenger_json["queue"],
                    p_processor_modules=passenger_json["processors"],
                    p_pusher_modules=passenger_json["pushers"],
                    p_sync_frequency=passenger_json["sync_frequency"],
                    p_queue_life_span=passenger_json["queue_life_span"])
                client_passengers.append(client_passenger)

            client_users = []
            if "users" in config_json:
                for user_json in config_json["users"]:
                    credential = Credential(
                        username=user_json["username"],
                        password=user_json["password"],
                        token=user_json["token"])
                    if "role" in user_json:
                        role = str_to_role(user_json["role"])
                    else:
                        role = Role.UNDEFINED
                    user = User(credential=credential, role=role)
                    client_users.append(user)

            client_obj = Client(
                p_id=client_directory,
                p_passengers=client_passengers,
                p_log_life_span=config_json["log_life_span"],
                p_users=client_users)

            output.append(client_obj)
        return output

    @property
    def client_directories(self) -> List[str]:
        """ Returns all client directories """
        client_root_path = JsonPathBuilder.get_client_root_path(self._args)
        return [f.name for f in scandir(client_root_path) if f.is_dir()]

    @property
    def client_master_as_json(self) -> str:
        """ Returns client master data as a JSON string """
        output_dict = {"client": []}
        for client in self.all_clients:
            client_dict = {
                "client_id": client.id,
                "log_life_span": client.log_life_span
            }
            output_dict["client"].append(client_dict)
        return json.dumps(output_dict, indent=4, sort_keys=True)

    @client_master_as_json.setter
    def client_master_as_json(self, p_master: str):
        """ Saves client master data from JSON string
        Check getter for JSON format
        """
        memory_dict = json.loads(p_master)
        database_dict = json.loads(self.client_master_as_json)
        table_keys = [TableKey("client", ["client_id"])]

        diff_check = DifferenceChecker(table_keys, memory_dict, database_dict)

        for diff_result in diff_check.result:
            if diff_result.action == Action.INSERT:
                self.create_client_if_missing(diff_result.row["client_id"], diff_result.row)
            elif diff_result.action == Action.UPDATE:
                self._save_config_dict(diff_result.row["client_id"], diff_result.row)
            elif diff_result.action == Action.DELETE:
                self._delete_client(diff_result.row["client_id"])

    def ensure_schema_existence(self, p_client_id: str):
        """ Checks the schema for the client
        Creates / completes the schema if anything is missing
        """
        self.create_client_if_missing(p_client_id, {})

    def get_config_as_json(self, p_client_id: str):
        """ Returns the contents of the config file as JSON """
        output = ""
        config_file_path = self._get_config_file_path(p_client_id)
        with open(config_file_path) as config_json_file:
            output = json.load(config_json_file)
        return output

    def get_config_as_string(self, p_client_id: str):
        """ Returns the contents of the config file as string """
        output = ""
        config_file_path = self._get_config_file_path(p_client_id)
        with open(config_file_path) as config_json_file:
            output = config_json_file.read()
        if "\r\n" in output:
            output = output.replace("\r\n", "\n")
        return output

    def get_single(self, p_id: str) -> Client:
        """ Returns a single client """
        if p_id == "" or p_id is None:
            raise ClientError(ClientError.ErrorCode.parameter_missing)
        for client in self.all_clients:
            if client.id == p_id:
                return client
        raise ClientError(ClientError.ErrorCode.client_not_found, p_client_id=p_id)

    def save_config(self, p_client_id: str, p_config: str):
        """ Writes the provided configuration to the disk """
        config_file_path = self._get_config_file_path(p_client_id)
        with open(config_file_path, "w") as config_json_file:
            config_json_file.write(p_config)

    def update_user_credential(self, p_client_id: str, p_credential: Credential):
        """ Updates the credential of a client user """
        for client_directory in self.client_directories:
            if client_directory != p_client_id:
                continue
            config_file_path = self._get_config_file_path(client_directory)
            with open(config_file_path) as config_json_file:
                config_json = json.load(config_json_file)
            if "users" not in config_json:
                config_json["users"] = []
            user_found = False
            for user in config_json["users"]:
                if user["username"] == p_credential.username:
                    user_found = True
                    user["password"] = p_credential.password
                    user["token"] = p_credential.token
            if not user_found:
                user = {
                    "username": p_credential.username,
                    "password": p_credential.password,
                    "token": p_credential.token
                }
                config_json["users"].append(user)
            with open(config_file_path, "w") as config_json_file:
                json.dump(config_json, config_json_file, indent=4, sort_keys=True)
            return

    def create_client_if_missing(self, p_client_id: str, p_base_dict: dict):
        """ Creates the client folder structure if not present """
        path_builder = JsonPathBuilder(p_client_id, self._args)
        pathlib.Path(path_builder.client_dir_path).mkdir(parents=True, exist_ok=True)
        pathlib.Path(path_builder.log_root_path).mkdir(parents=True, exist_ok=True)
        pathlib.Path(path_builder.queue_root_path).mkdir(parents=True, exist_ok=True)
        if not path.exists(path_builder.config_file_path):
            self._save_config_dict(p_client_id, p_base_dict)

    def create_client_if_missing__by_obj(self, p_client: Client):
        """ Creates the client folder structure if not present """

        client_dict = {
            "log_life_span": p_client.log_life_span,
            "passengers": [],
            "users": []
        }

        for passenger in p_client.passengers:
            passenger_dict = {
                "name": passenger.name,
                "processors": copy(passenger.processor_modules),
                "pullers": copy(passenger.puller_modules),
                "pushers": copy(passenger.pusher_modules),
                "queue": passenger.queue_module,
                "queue_life_span": passenger.queue_life_span,
                "sync_frequency": passenger.sync_frequency
            }
            client_dict["passengers"].append(passenger_dict)

        for user in p_client.users:
            user_dict = {
                "password": user.credential.password,
                "token": user.credential.token,
                "username": user.credential.username
            }
            client_dict["users"].append(user_dict)

        self.create_client_if_missing(p_client.id, client_dict)

    def _delete_client(self, p_client_id: str):
        path_builder = JsonPathBuilder(p_client_id, self._args)
        rmtree(path_builder.client_dir_path)

    def _get_config_file_path(self, p_client_id: str) -> str:
        path_builder = JsonPathBuilder(p_client_id, self._args)
        return path_builder.config_file_path

    def _save_config_dict(self, p_client_id: str, p_base_dict: dict):
        config = copy(p_base_dict)
        if "client_id" in config:
            del config["client_id"]
        if "log_life_span" not in config:
            config["log_life_span"] = 1
        if "passengers" not in config:
            config["passengers"] = []
        if "users" not in config:
            config["users"] = []

        self.save_config(p_client_id, json.dumps(config))



""" Default processor factory module """
from vibhaga.inspector import Inspector
from databus.client.log import Log
from databus.processor.abstract_factory import AbstractProcessorFactory, ProcessorCreationError
from databus.processor.abstract_processor import AbstractProcessor


class PrimalProcessorFactory(AbstractProcessorFactory): # pylint: disable=R0903
    """ Default processor factory class """

    def create_processor(self, p_module: str, p_log: Log) -> AbstractProcessor:
        """ Factory method """
        if p_module == "" or p_module is None:
            raise ProcessorCreationError(ProcessorCreationError.ErrorCode.parameter_missing)

        candidates = Inspector.get_classes_in_module(
            p_module,
            exclude_classes=["AbstractProcessor"])

        for candidate in candidates:
            try:
                obj_instance = candidate(p_log)
                if isinstance(obj_instance, AbstractProcessor):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                pass

        raise ProcessorCreationError(
            ProcessorCreationError.ErrorCode.cant_create_instance,
            p_module=p_module)



""" Default dispatcher module """
from datetime import datetime, timedelta
from threading import Thread
from time import sleep
from typing import List
import os
from databus.client.client import Client
from databus.client.client_passenger import ClientPassenger
from databus.client.log import Log, LogEntry, MessageType
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher, DispatcherTicket
from databus.driver.abstract_driver import BusTicket
from databus.web import app


class ClientPassengerTickCount:
    """ Keeps track of passenger ticks """
    _INFINITE_TICK = 999999999

    def __init__(self):
        self._ticks = {}

    def collect(self, p_client_id: str, p_passenger_id: str):
        """ Inserts the client & passenger if they weren't before """
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        if key not in self._ticks:
            self._ticks[key] = 0

    def collect_clients(self, p_clients: List[Client]):
        """ Inserts missing clients """
        for client in p_clients:
            for passenger in client.passengers:
                self.collect(client.id, passenger.name)

    def expedite_client_passenger(self, p_client_id: str, p_passenger_id: str):
        """ Prioritizes the passenger in the next cycle """
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        if key not in self._ticks:
            return
        self._ticks[key] = ClientPassengerTickCount._INFINITE_TICK

    def get_tick(self, p_client_id: str, p_passenger_id: str):
        """ Returns the current tick count """
        self.collect(p_client_id, p_passenger_id)
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        return self._ticks[key]

    def reset_tick(self, p_client_id: str, p_passenger_id: str):
        """ Resets tick """
        self.collect(p_client_id, p_passenger_id)
        key = ClientPassengerTickCount._build_key(p_client_id, p_passenger_id)
        self._ticks[key] = 0

    def tick(self):
        """ Increases tick counts for all client passengers """
        for self_tick in self._ticks:
            self._ticks[self_tick] += 1

    @staticmethod
    def _build_key(p_client_id: str, p_passenger_id: str):
        return p_client_id + ":-:" + p_passenger_id


class DispatchState: # pylint: disable=R0903
    """ State of the dispatcher """
    def __init__(self):
        self.clients = List[Client]


class PrimalDispatcher(AbstractDispatcher): # pylint: disable=R0903
    """ Default dispatcher implementation """

    def __init__(self, p_ticket: DispatcherTicket = None):
        super().__init__(p_ticket)
        self._dispatch_state = DispatchState()
        self._next_dispatch_time = datetime.now()
        self._tick_count = ClientPassengerTickCount()
        self._paused = False
        self._dispatching = False
        self._shutting_down = False

    def expedite_client_passenger(self, p_client_id: str, p_passenger_module: str):
        """ Prioritizes the passenger in the next cycle """
        self._tick_count.expedite_client_passenger(p_client_id, p_passenger_module)

    def export_data_begin(self):
        """ Indicates that data export is starting """
        if self._dispatching:
            raise Exception("Can't export while dispatching")
        if self._shutting_down:
            raise Exception("Can't export while shutting down")
        self._paused = True

    def export_data_end(self):
        """ Indicates that data export is ending """
        self._paused = False

    def start(self):
        """ Starts the dispatcher timer and web server """
        if self.ticket.run_web_server:
            Thread(target=self._start_web_server, daemon=True).start()
        self._start_dispatch_timer()

    def request_shutdown(self):
        """ Dispatcher shutdown """
        self._shutting_down = True
        Thread(target=self._shutdown_when_safely_possible, daemon=True).start()

    def _dispatch(self):
        try:
            if self._paused or self._shutting_down:
                return
            self._dispatching = True
            self._dispatch_state = DispatchState()
            self._read_clients()
            self._drive_high_time_passengers()
        finally:
            self._dispatching = False

    def _drive_high_time_passengers(self):
        for client in self._dispatch_state.clients:
            for client_passenger in client.passengers:
                try:
                    tick_count = self._tick_count.get_tick(client.id, client_passenger.name)
                    if tick_count < client_passenger.sync_frequency:
                        continue
                    self._tick_count.reset_tick(client.id, client_passenger.name)
                    self._drive_passenger(client, client_passenger)
                except Exception as drive_error: # pylint: disable=W0703
                    print(str(drive_error))

    def _drive_passenger(self, p_client: Client, p_client_passenger: ClientPassenger):
        if self._shutting_down:
            return
        log = Log()
        db = None # pylint: disable=C0103
        driver = None
        try:
            log.append_text("Dispatching client " +
                            p_client.id +
                            " passenger " +
                            p_client_passenger.name)

            log.append_text("Creating database " + self.ticket.database_module)
            db = self.get_client_database(p_client.id, log) # pylint: disable=C0103
            log.append_text("Checking schema")
            db.ensure_schema_existence()
            log.append_text("Creating driver " + self.ticket.driver_module)
            driver = self.get_driver()
            log.append_text("Creating ticket")

            ticket = BusTicket(
                p_client_passenger=p_client_passenger,
                p_log=log,
                p_database=db)

            log.append_text("Driving")
            driver.drive(ticket)

        except Exception as drive_error: # pylint: disable=W0703
            if log is not None:
                log.append_entry(LogEntry(p_message=str(drive_error), p_type=MessageType.error))
        finally:
            self._tick_count.reset_tick(p_client.id, p_client_passenger.name)
            if self.ticket.dispatcher_observer is not None and log is not None:
                self.ticket.dispatcher_observer.drive_passenger_complete(p_client,
                                                                         p_client_passenger,
                                                                         log)
            if db is not None:
                db.insert_log(log)
                db.delete_old_logs(p_client.log_expiry_date)
            if driver is not None:
                driver.queue.delete_completed_passengers(
                    p_client_passenger.name,
                    p_client_passenger.queue_expiry_date)

    def _read_clients(self):
        self._dispatch_state.clients = self.all_clients
        self._tick_count.collect_clients(self._dispatch_state.clients)
        self._tick_count.tick()

    def _sleep_until_next_dispatch_time(self):
        now = datetime.now()
        if now >= self._next_dispatch_time:
            return
        seconds_to_sleep = (self._next_dispatch_time - now).seconds
        sleep(seconds_to_sleep)

    def _start_dispatch_timer(self):
        while True:
            self._next_dispatch_time = self._next_dispatch_time + timedelta(0, 60)
            self._dispatch()
            if self._shutting_down:
                return
            self._sleep_until_next_dispatch_time()

    def _start_web_server(self):
        app.run_web_server(self)

    def _shutdown_when_safely_possible(self):
        while True:
            if not self._dispatching and not self._paused:
                os._exit(0) # pylint: disable=W0212
                return
            sleep(1)



""" Queue status module """
from enum import Enum
from typing import List
from databus.client.client_passenger import ClientPassenger
from databus.passenger.abstract_passenger import AbstractPassenger



class QueueStatus(Enum):
    """ Queue status codes """
    undefined = 0
    incomplete = 1
    complete = 2

class ProcessorQueueStatus: # pylint: disable=R0903
    """ Processor queue status class """
    def __init__(self, p_processor_module: str, p_status: QueueStatus = QueueStatus.undefined):
        self.processor_module = p_processor_module
        self.status = p_status

class PusherQueueStatus: # pylint: disable=R0903
    """ Pusher queue status class """
    def __init__(self, p_pusher_module: str, p_status: QueueStatus = QueueStatus.undefined):
        self.pusher_module = p_pusher_module
        self.status = p_status

class PassengerQueueStatus:
    """ Passenger queue status class """
    def __init__(self,
                 p_passenger: AbstractPassenger,
                 p_pusher_statuses: List[PusherQueueStatus] = None,
                 p_processor_statuses: List[ProcessorQueueStatus] = None,
                 p_puller_notified: bool = False):

        self.passenger = p_passenger

        if p_pusher_statuses is None:
            self.pusher_statuses = []
        else:
            self.pusher_statuses = p_pusher_statuses

        if p_processor_statuses is None:
            self.processor_statuses = []
        else:
            self.processor_statuses = p_processor_statuses

        self.puller_notified = p_puller_notified

    @property
    def all_processors_complete(self) -> bool:
        """ Returns true if all processors are complete """
        for processor_status in self.processor_statuses:
            if processor_status.status != QueueStatus.complete:
                return False
        return True

    @property
    def all_pushers_complete(self) -> bool:
        """ Returns true if all pushers are complete """
        for pusher_status in self.pusher_statuses:
            if pusher_status.status != QueueStatus.complete:
                return False
        return True

    @property
    def status(self) -> QueueStatus:
        """ Returns the overall status of the item """
        if self.puller_notified and self.all_processors_complete and self.all_pushers_complete:
            return QueueStatus.complete
        return QueueStatus.incomplete

    def set_all_processor_statuses(self, p_status: QueueStatus):
        """ Updates statuses of all processors """
        for processor_status in self.processor_statuses:
            processor_status.status = p_status

    def set_all_pusher_statuses(self, p_status: QueueStatus):
        """ Updates statuses of all pushers """
        for pusher_status in self.pusher_statuses:
            pusher_status.status = p_status

    def set_processor_status(self, p_processor_module: str, p_status: QueueStatus):
        """ Updates status of a single processor """
        for processor_status in self.processor_statuses:
            if processor_status.processor_module == p_processor_module:
                processor_status.status = p_status

    def set_pusher_status(self, p_pusher_module: str, p_status: QueueStatus):
        """ Updates status of a single pusher """
        for pusher_status in self.pusher_statuses:
            if pusher_status.pusher_module == p_pusher_module:
                pusher_status.status = p_status


class QueueStatusFactory: # pylint: disable=R0903
    """ Queue status factory class """
    @staticmethod
    def get_passenger_queue_status(p_passenger: AbstractPassenger,
                                   p_client_passenger: ClientPassenger
                                   ) -> PassengerQueueStatus:
        """ Returns the queue status of a single passenger """

        output = PassengerQueueStatus(p_passenger)

        for processor_module in p_client_passenger.processor_modules:
            output.processor_statuses.append(
                ProcessorQueueStatus(processor_module, QueueStatus.incomplete))

        for pusher_module in p_client_passenger.pusher_modules:
            output.pusher_statuses.append(
                PusherQueueStatus(pusher_module, QueueStatus.incomplete))

        return output


class QueueStatusError(Exception):
    """ Queue status exception class """

    class ErrorCode(Enum):
        """ Queue status error code """
        module_missing = 1

    def __init__(self,
                 p_error_code: ErrorCode,
                 p_queue_status_type: str,
                 p_passenger_id: str):
        super().__init__()
        self.error_code = p_error_code
        self.queue_status_type = p_queue_status_type
        self.passenger_id = p_passenger_id

    @property
    def message(self) -> str:
        """ Returns error message as string """
        if self.error_code == QueueStatusError.ErrorCode.module_missing:
            return "Passenger " + self.passenger_id + " is missing " + self.queue_status_type + " module" # pylint: disable=C0301
        return "Passenger error"


class Validator: # pylint: disable=R0903
    """ Queue module validator class """
    @staticmethod
    def validate_queue_module(p_queue_status_type: str,
                              p_passenger_id: str,
                              p_module: str):
        """ Validates queue module """
        if p_module is None or p_module == "":
            raise QueueStatusError(p_passenger_id=p_passenger_id,
                                   p_error_code=QueueStatusError.ErrorCode.module_missing,
                                   p_queue_status_type=p_queue_status_type)



""" Databus initialization module """
from os import path
from sys import modules

__version__ = "6.8.0"
AUTHOR = "Kerem Koseoglu"
EMAIL = "kerem@keremkoseoglu.com"
DESCRIPTION = "Databus is a framework to transfer data between systems."
PYTHON_VERSION = ">=3.6.5"


def get_root_path() -> str:
    """ Returns the root path of Databus """
    return path.dirname(modules['__main__'].__file__)



""" Module for external configuration files """
from typing import List


class ExternalConfigFile: # pylint: disable=R0903
    """ External configuration file """
    def __init__(self, p_client_id: str, p_file_id: str, p_path: str):
        self.client_id = p_client_id
        self.file_id = p_file_id
        self.path = p_path

    @property
    def file_content(self) -> str:
        """ Reads file content from the disk & returns """
        with open(self.path, "r") as text_file:
            output = text_file.read()
        return output

    @file_content.setter
    def file_content(self, p_content: str):
        """ Writes file content to disk """
        with open(self.path, "w") as text_file:
            text_file.write(p_content)


class ExternalConfigFileManager:
    """ Manager of external configuration files """
    def __init__(self):
        self._files = {}

    def add_file(self, p_client_id: str, p_file_id: str, p_path: str):
        """ Registers a new file """
        file_key = ExternalConfigFileManager._build_file_key(p_client_id, p_file_id)
        file_obj = ExternalConfigFile(p_client_id, p_file_id, p_path)
        self._files[file_key] = file_obj

    def add_files(self, p_files: List[ExternalConfigFile]):
        """ Registers multiple files """
        for file in p_files:
            file_key = ExternalConfigFileManager._build_file_key(file.client_id, file.file_id)
            self._files[file_key] = file

    def get_file(self, p_client_id: str, p_file_id: str) -> ExternalConfigFile:
        """ Returns an existing file """
        file_key = ExternalConfigFileManager._build_file_key(p_client_id, p_file_id)
        return self._files[file_key]

    def get_files_of_client(self, p_client_id: str) -> List[ExternalConfigFile]:
        """ Returns all files of given client """
        output = []
        for file_key in self._files:
            file_obj = self._files[file_key]
            if file_obj.client_id == p_client_id:
                output.append(file_obj)
        return output

    @staticmethod
    def _build_file_key(p_client_id: str, p_file_id: str) -> str:
        return p_client_id + "__" + p_file_id



""" Demo pusher module """
from databus.pqueue.queue_status import PassengerQueueStatus, QueueStatus
from databus.pusher.abstract_pusher import AbstractPusher


class DemoPusher2(AbstractPusher): # pylint: disable=R0903
    """ Demo pusher class """
    def push(self, p_passenger: PassengerQueueStatus):
        """ Push demo """
        self.log.append_text("Pushed passenger " + p_passenger.passenger.id_text)
        p_passenger.set_pusher_status(self.__module__, QueueStatus.complete)



""" Default driver factory module """
from vibhaga.inspector import Inspector
from databus.driver.abstract_driver import AbstractDriver
from databus.driver.abstract_factory import AbstractDriverFactory, DriverCreationError
from databus.pqueue.abstract_factory import AbstractQueueFactory
from databus.processor.abstract_factory import AbstractProcessorFactory
from databus.puller.abstract_factory import AbstractPullerFactory
from databus.pusher.abstract_factory import AbstractPusherFactory


class PrimalDriverFactory(AbstractDriverFactory): # pylint: disable=R0903
    """ Default driver factory class """

    def create_driver(self, # pylint: disable=R0913
                      p_module: str,
                      p_queue_factory: AbstractQueueFactory,
                      p_processor_factory: AbstractProcessorFactory,
                      p_puller_factory: AbstractPullerFactory,
                      p_pusher_factory: AbstractPusherFactory
                      ) -> AbstractDriver:
        """ Creates a new driver object """
        if p_module == "" or p_module is None:
            raise DriverCreationError(DriverCreationError.ErrorCode.parameter_missing)

        candidates = Inspector.get_classes_in_module(p_module, exclude_classes=["AbstractDriver"])
        for candidate in candidates:
            try:
                obj_instance = candidate(
                    p_queue_factory,
                    p_processor_factory,
                    p_puller_factory,
                    p_pusher_factory)
                if isinstance(obj_instance, AbstractDriver):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                pass

        raise DriverCreationError(
            DriverCreationError.ErrorCode.cant_create_instance,
            p_module=p_module)



""" Module to help with SQL queries """
from typing import List
from databus.database.sql_db.delete_builder import DeleteBuilder
from databus.database.sql_db.driver.pyodbc_driver import PyodbcDriver
from databus.database.sql_db.insert_builder import InsertBuilder
from databus.database.sql_db.path_builder import PathBuilder
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments
from databus.database.sql_db.update_builder import UpdateBuilder
from databus.database.sql_db.where_builder import WhereBuilder

class QueryHelper:
    """ Helper class to access SQL server easier """
    def __init__(self, p_arguments: dict, p_client_id: str, p_auto_commit: bool = False):
        self.args = SqlDatabaseArguments(p_arguments)
        self._client_id = p_client_id
        self._driver = PyodbcDriver()
        self._driver.autocommit = p_auto_commit
        self._driver.connect(SqlDatabaseArguments(p_arguments))
        self._where = WhereBuilder(p_client_id=self._client_id)
        self.path_builder = PathBuilder(self.args)

    @property
    def autocommit(self) -> bool:
        """ Are commands committed automatically """
        return self._driver.autocommit

    @property
    def client_id(self) -> str:
        """ Client ID """
        return self._client_id

    @autocommit.setter
    def autocommit(self, p_active: bool):
        """ Are commands committed automatically """
        self._driver.autocommit = p_active

    def commit(self):
        """ Runs a commit operation via the driver """
        self._driver.commit()

    def delete(self, p_table: str, p_where: str = ""):
        """ Deletes entries from SQL server """
        command = "DELETE FROM "
        command += self.path_builder.get_table_path(p_table)
        command += self._where.build(p_where)
        self._driver.execute_sql(command)

    def execute_delete(self, p_delete: DeleteBuilder):
        """ Executes a delete statement """
        self._driver.execute_sql(p_delete.delete_command)

    def execute_insert(self, p_insert: InsertBuilder):
        """ Executes an Insert statement """
        self._driver.execute_sql(p_insert.insert_command)

    def execute_stored_procedure(self, p_sql: str, p_values):
        """ Executes a stored procedure """
        self._driver.execute_stored_procedure(p_sql, p_values)

    def execute_update(self, p_update: UpdateBuilder):
        """ Executes an Insert statement """
        self._driver.execute_sql(p_update.update_command)

    def rollback(self):
        """ Runs a rollback operation via the driver """
        self._driver.rollback()

    def select_all(self,
                   p_table: str,
                   p_where: str = "",
                   p_order_fields: List[str] = None) -> List[dict]:
        """ Selects & returns all entries from table
        The where condition will be touched - client id will be added automatically
        """
        where = self._where.build(p_where, p_order_fields=p_order_fields)
        return self.select_all_literal_where(p_table, p_literal_where=where)

    def select_all_literal_where(self, p_table: str, p_literal_where: str = "") -> List[dict]:
        """ Selects & returns all entries from table
        The where condition is used as a literal value, so it's not polluted by
        client ID or anything.
        """
        query = "SELECT * FROM " + self.path_builder.get_table_path(p_table) + p_literal_where
        return self._driver.select(query)

    def select_all_no_where(self, p_table: str, p_order_by: str = "") -> dict:
        """ Selects & returns all entries from table, without WHERE conditions """
        query = "SELECT * FROM " + self.path_builder.get_table_path(p_table)
        if p_order_by != "":
            query += " ORDER BY " + p_order_by
        return self._driver.select(query)

    def select_all_where_builder(self, p_table: str, p_builder: WhereBuilder) -> dict:
        """ Selects & returns all entries from table which match the builder """
        return self.select_all_literal_where(p_table, p_builder.where)

    def select_single(self, p_table: str, p_where: str = "") -> dict:
        """ Selects & returns a single entry
        The where condition will be touched - client id will be added automatically
        """
        where = self._where.build(p_where)
        return self.select_all_literal_where(p_table, p_literal_where=where)[0]



""" Controller module dealing with logging in and out """
from flask import make_response, redirect, request, session, url_for
from databus.client.client import Credential
from databus.client.user import Role
from databus.web.controller.abstract_controller import AbstractController


class LoginAttemptController(AbstractController):
    """ Login attempt """

    def execute(self):
        """ Builds and returns the page """
        try:
            client_id = request.form["client_id"]
            username = request.form["username"]
            password = request.form["password"]
            remember = False
            if "remember" in request.form:
                if request.form["remember"] == "X":
                    remember = True

            client_db = self.dispatcher.get_client_database(client_id)
            credential = None
            if client_db.client.authorization_active:
                credential = Credential(username=username, password=password)
                user = client_db.client.authenticate(credential)
                if user is None:
                    return redirect(url_for("_login"), code=302)
                user_role = user.role
                if remember:
                    credential.generate_token()
                    client_db.update_user_credential(credential)
            else:
                user_role = Role.ADMINISTRATOR

            session["client_id"] = client_db.client_id
            session["user_role"] = user_role.name
            resp = make_response(redirect(url_for("_home"), code=302))
            if credential is not None and credential.token != "":
                resp.set_cookie("token", credential.token)
            return resp

        except Exception: # pylint: disable=W0703
            return redirect(url_for("_login"), code=302)


class LogoffController(AbstractController):
    """ Logoff """

    def execute(self):
        """ Builds and returns the page """
        session["client_id"] = ""
        resp = make_response(redirect(url_for("_login"), code=302))
        resp.set_cookie("token", "")
        return resp



""" Module to build insert conditions """
from databus.database.sql_db.modifiable import Modifiable


class InsertBuilder(Modifiable):
    """ Helper class to build insert commands """

    @property
    def insert_command(self) -> str:
        """ Insert command """
        query = "INSERT INTO " + self.table + " (client_id "

        for keyval in self.key_values:
            query += " , " + keyval.key

        query += " ) VALUES ( '" + self._client_id + "' "

        for keyval in self.key_values:
            query += " , "
            if keyval.is_string:
                query += "'"
            query += Modifiable._get_safe_string(keyval.val)
            if keyval.is_string:
                query += "'"

        query += ")"

        return query



""" Databus setup module """
import os
import setuptools
import databus


def get_databus_requirements() -> []:
    """ Returns a list of requirements """
    output = []
    lib_folder = os.path.dirname(os.path.realpath(__file__))
    requirement_path = lib_folder + '/requirements.txt'

    if os.path.isfile(requirement_path):
        with open(requirement_path) as f:
            output = f.read().splitlines()

    return output


setuptools.setup(
    name="databus-keremkoseoglu",
    version=databus.__version__,
    author=databus.AUTHOR,
    author_email=databus.EMAIL,
    description=databus.DESCRIPTION,
    long_description="A framework to transfer data between systems",
    long_description_content_type="text/markdown",
    url="https://github.com/keremkoseoglu/databus",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=databus.PYTHON_VERSION,
    install_requires=[
        "cython",
        "flask",
        "pyodbc",
        "urlextract",
        "vibhaga-keremkoseoglu",
        "waitress"
    ],
    include_package_data=True
)



""" Default database factory module """
from typing import List
from vibhaga.inspector import Inspector
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.database.abstract_factory import AbstractDatabaseFactory, DatabaseCreationError
from databus.database.json_db.json_database import JsonDatabase # pylint: disable=W0611
from databus.database.sql_db.sql_database import SqlDatabase # pylint: disable=W0611
from databus.passenger.abstract_factory import AbstractPassengerFactory


class PrimalDatabaseFactory(AbstractDatabaseFactory): # pylint: disable=R0903
    """ Default database factory class """
    def create_database(self, # pylint: disable=R0913
                        p_module: str,
                        p_client_id: str,
                        p_log: Log,
                        p_passenger_factory: AbstractPassengerFactory,
                        p_arguments: dict
                        ) -> AbstractDatabase:
        """ Default method for database creation """
        if p_module == "" or p_module is None:
            raise DatabaseCreationError(DatabaseCreationError.ErrorCode.parameter_missing)

        if p_client_id is None:
            client_id = ""
        else:
            client_id = p_client_id

        candidates = Inspector.get_classes_in_module(
            p_module,
            exclude_classes=["AbstractDatabase", "datetime"])

        for candidate in candidates:
            try:
                obj_instance = candidate(client_id, p_log, p_passenger_factory, p_arguments)
                if isinstance(obj_instance, AbstractDatabase):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                continue

        raise DatabaseCreationError(DatabaseCreationError.ErrorCode.cant_create_instance,
                                    p_module=p_module,
                                    p_client_id=p_client_id)

    @property
    def database_modules(self) -> List[str]:
        """ Returns a list of database modules in the system """
        output = []
        for subclass in AbstractDatabase.__subclasses__():
            output.append(subclass.__module__)
        return output

    @property
    def database_classes(self) -> List[str]:
        """ Returns a list of database classes in the system """
        return ["JsonDatabase", "SqlDatabase"]



""" JSON log module """
from datetime import datetime
from os import path, remove, scandir
from typing import List
from databus.client.log import Log
from databus.database.json_db.json_client import JsonClient
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.json_db.json_path_builder import JsonPathBuilder


class JsonLog:
    """ JSON log class """
    def __init__(self, args: JsonDatabaseArguments):
        self._args = args
        self._client = JsonClient(args)

    def build_log_file_name(self, p_log: Log) -> str:
        """ Builds log file name """
        datetime_part = p_log.creation_datetime.isoformat()
        safe_datetime_part = datetime_part.replace(":", "_")
        guid_part = str(p_log.guid)
        return safe_datetime_part + "_" + guid_part + "." + self._args.log_extension

    def build_log_file_path(self, p_client_id: str, p_log: Log) -> str:
        """ Builds log file path """
        return path.join(self.get_root_path(p_client_id),
                         self.build_log_file_name(p_log))

    def delete_log_file_before(self, p_client_id: str, p_before: datetime, p_log: Log):
        """ Deletes log files before the given date """
        log_root_path = self.get_root_path(p_client_id)
        all_log_files = self.get_log_file_list(p_client_id)
        for log_file in all_log_files:
            split1 = log_file.split("T")
            split2 = split1[0].split("-")
            log_file_date = datetime(year=int(split2[0]), month=int(split2[1]), day=int(split2[2]))
            if log_file_date < p_before:
                full_log_file_path = path.join(log_root_path, log_file)
                p_log.append_text("Deleting " + full_log_file_path)
                remove(full_log_file_path)

    def get_log_file_content(self, p_client_id: str, p_log_file: str) -> str:
        """ Returns the content of the given log file """
        output = ""
        log_path = path.join(self.get_root_path(p_client_id), p_log_file)
        with open(log_path, mode="r") as log_file:
            output = log_file.read()
        return output

    def get_log_file_list(self, p_client_id: str) -> List[str]:
        """ Log file list """
        output = []
        log_root_path = self.get_root_path(p_client_id)
        file_list = [f.name for f in scandir(log_root_path) if f.is_file()]
        supposed_extension = self._args.log_extension.lower()
        for file_candidate in file_list:
            extension = path.splitext(file_candidate)[1].replace(".", "").lower()
            if extension == supposed_extension:
                output.append(file_candidate)
        return output

    def get_root_path(self, p_client_id: str) -> str:
        """ Returns the root log path for the given client """
        return self._get_path_builder(p_client_id).log_root_path

    def insert(self, p_client_id: str, p_log: Log):
        """ Writes log file to disk """
        log_file_content = p_log.entries_as_string
        log_file_path = self.build_log_file_path(p_client_id, p_log)

        log_file = open(log_file_path, "w+")
        log_file.write(log_file_content)
        log_file.close()

    def _get_path_builder(self, p_client_id: str) -> JsonPathBuilder:
        return JsonPathBuilder(p_client_id, self._args)



""" Abstract passenger factory module """
from abc import ABC, abstractmethod
from enum import Enum
from databus.passenger.abstract_passenger import AbstractPassenger


class PassengerCreationError(Exception):
    """ Passenger creation exception class """
    class ErrorCode(Enum):
        """ Passenger creation error codes """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self, p_error_code: ErrorCode, p_module: str = ""):
        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

    @property
    def message(self) -> str:
        """ Passenger creation error message """
        if self.error_code == PassengerCreationError.ErrorCode.cant_create_instance:
            return "Can't create " + self.module + " passenger instance"
        if self.error_code == PassengerCreationError.ErrorCode.parameter_missing:
            return "Parameters missing, can't create passenger instance"
        return "Passenger creation error"


class AbstractPassengerFactory(ABC): # pylint: disable=R0903
    """ Abstract passenger factory class """
    @abstractmethod
    def create_passenger(self, p_module: str) -> AbstractPassenger:
        """ Abstract passenger factory """



""" Abstract database module """
from abc import ABC, abstractmethod
from datetime import datetime
from typing import List
from uuid import UUID
from databus.client.client import Client
from databus.client.log import Log, MessageType
from databus.client.user import Credential
from databus.passenger.abstract_factory import AbstractPassengerFactory
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.pqueue.queue_status import PassengerQueueStatus, QueueStatus


class LogListItem: # pylint: disable=R0903
    """ A structure for a log list item """
    def __init__(self,
                 p_log_id: str = "",
                 p_worst_message_type: MessageType = MessageType.info):
        self.log_id = p_log_id
        self.worst_message_type = p_worst_message_type


class AbstractDatabase(ABC):
    """ Abstract database class """
    def __init__(self,
                 p_client_id: str,
                 p_log: Log,
                 p_passenger_factory: AbstractPassengerFactory,
                 p_arguments: dict):
        self.client_id = p_client_id
        self.log = p_log
        self.passenger_factory = p_passenger_factory
        self.arguments = p_arguments
        self.client = Client()

    @property
    @abstractmethod
    def client_master_data(self) -> str:
        """ Returns the definition of all clients as a string
        This string (preferably JSON) will be shown to the
        root user as a JSON file.
        When edited, it will be set back
        """

    @client_master_data.setter
    @abstractmethod
    def client_master_data(self, p_definitions: str):
        """ Sets the definitions returned from the GUI
        Subclasses are expected to write those definitions
        to the disk, database, etc
        """

    @property
    @abstractmethod
    def customizing(self) -> str:
        """ Returns the client customizing in text format
        (preferably JSON)
        """

    @customizing.setter
    @abstractmethod
    def customizing(self, p_customizing: str):
        """ Sets the customizing returned from the GUI
        Subclasses are expected to write those settings
        to the disk, database, etc
        """

    @abstractmethod
    def delete_old_logs(self, p_before: datetime):
        """ Deletes overdue logs """

    @abstractmethod
    def delete_passenger_queue(self, p_passengers: List[AbstractPassenger]):
        """ Deletes the given passengers from queue """

    @abstractmethod
    def ensure_schema_existence(self):
        """ Checks the schema for the client
        Creates / completes the schema if anything is missing
        """

    @abstractmethod
    def erase_passenger_queue(self):
        """ Eradicates queue completely """

    @abstractmethod
    def get_clients(self) -> List[Client]:
        """ Returns clients """

    @abstractmethod
    def get_log_content(self, p_log_id: str) -> str:
        """ Returns the contents of the given log
        p_log_id is whatever you have returned in get_log_list.
        """

    @abstractmethod
    def get_log_list(self) -> List[LogListItem]:
        """ Returns a list of log files """

    @abstractmethod
    def get_passenger_queue_entries(self, # pylint: disable=R0913
                                    p_passenger_module: str = None,
                                    p_processor_status: QueueStatus = None,
                                    p_pusher_status: QueueStatus = None,
                                    p_puller_notified: bool = None,
                                    p_pulled_before: datetime = None
                                    ) -> List[PassengerQueueStatus]:
        """ Returns requested passenger queue entries """

    @abstractmethod
    def get_passenger_queue_entry(self, # pylint: disable=R0913
                                  p_internal_id: str
                                 ) -> PassengerQueueStatus:
        """ Returns requested passenger queue entry """

    @abstractmethod
    def insert_client(self, p_client: Client):
        """ Inserts a new client """

    @abstractmethod
    def insert_log(self, p_log: Log):
        """ Inserts new log entries """

    @abstractmethod
    def insert_passenger_queue(self, p_passenger_status: PassengerQueueStatus):
        """ Adds the to queue """

    @abstractmethod
    def update_queue_status(self, p_status: PassengerQueueStatus):
        """ Updates the status of the passenger in the queue """

    @abstractmethod
    def update_user_credential(self, p_credential: Credential):
        """ Updates the credential of the given user """

    @abstractmethod
    def convert_log_guid_to_id(self, p_guid: UUID) -> str:
        """ UUID to id conversion """

    @abstractmethod
    def _get_client(self, p_id: str) -> Client:
        """ Returns a new client """



""" Module for modifiable key - values """
import binascii
from datetime import datetime
from typing import List
from databus.database.sql_db.path_builder import PathBuilder
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments
from databus.database.sql_db.value_conversion import DatabusToSql


class KeyValue: # pylint: disable=R0903
    """ Key - Value to be used in modification """
    def __init__(self, p_key: str, p_val: str, p_is_string: bool):
        self.key = p_key
        self.val = p_val
        self.is_string = p_is_string


class Modifiable:
    """ Class for modifiable key - value pairs """
    def __init__(self, p_args: SqlDatabaseArguments, p_client_id: str):
        self._client_id = p_client_id
        self._key_values = []
        self._table = ""
        self._args = p_args
        self._path_builder = PathBuilder(p_args)
        self.clear()

    @property
    def key_values(self) -> List[KeyValue]:
        """ Returns the output """
        return self._key_values

    @property
    def table(self) -> str:
        """ Table to be inserted """
        return self._path_builder.get_table_path(self._table)

    @table.setter
    def table(self, p_table: str):
        """ Table setter """
        self.clear()
        self._table = p_table

    def add_binary(self, p_key: str, p_bin: bytearray):
        """ Adds binary content """
        filecontent_hex = '0x'.encode('ascii') + binascii.hexlify(p_bin)
        self.add_string(p_key, filecontent_hex)

    def add_datetime(self, p_key: str, p_val: datetime):
        """ Adds a new datetime """
        keyval = KeyValue(p_key, DatabusToSql.date_time(p_val), True)
        self._key_values.append(keyval)

    def add_int(self, p_key: str, p_val: int):
        """ Adds a new integer """
        keyval = KeyValue(p_key, str(p_val), False)
        self._key_values.append(keyval)

    def add_string(self, p_key: str, p_val):
        """ Adds a new string """
        keyval = KeyValue(p_key, str(p_val), True)
        self._key_values.append(keyval)

    def add_value(self, p_key_value: KeyValue):
        """ Adds a new key - value pair """
        self._key_values.append(p_key_value)

    def clear(self):
        """ Reset """
        self._key_values = []
        self._table = ""

    @staticmethod
    def _get_safe_string(p_input: str) -> str:
        return "{}".format(p_input).replace("'", "''")



""" Module for SQL databases

Command to start docker module:
docker run -d --name sql_server_demo -e 'ACCEPT_EULA=Y' -e 'SA_PASSWORD=reallyStrongPwd123' -p 1433:1433 microsoft/mssql-server-linux
Check creation_script.sql to create a new database
"""
from datetime import datetime
import json
from typing import List
from uuid import UUID
from databus.client.client import Client, ClientPassenger
from databus.client.log import Log
from databus.client.user import Credential, User, Role, str_to_role
from databus.database.abstract_database import AbstractDatabase, LogListItem
from databus.passenger.abstract_factory import AbstractPassengerFactory
from databus.passenger.abstract_passenger import AbstractPassenger, Attachment
from databus.pqueue.queue_status import \
    QueueStatus, PassengerQueueStatus, ProcessorQueueStatus, PusherQueueStatus
from databus.database.sql_db.action_decider import ActionDecider
from databus.database.sql_db.insert_builder import InsertBuilder
from databus.database.sql_db.query_helper import QueryHelper
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments
from databus.database.sql_db.update_builder import UpdateBuilder
from databus.database.sql_db.value_conversion import DatabusToSql, SqlToDatabus
from databus.database.sql_db.where_builder import WhereBuilder


ARGS_TEMPLATE = SqlDatabaseArguments.TEMPLATE


class ClientDataset:
    """ Defines a client dataset """
    def __init__(self, p_query_helper: QueryHelper):
        self._query_helper = p_query_helper
        self.client_list = []
        self.passenger_list = []
        self.puller_list = []
        self.processor_list = []
        self.pusher_list = []
        self.user_list = []

    def dump_as_client_list(self) -> List[Client]:
        """ Returns output as a client list """
        output = []

        for client_row in self.client_list:
            client = Client(p_id=client_row["client_id"],
                            p_log_life_span=client_row["log_life_span"])
            for passenger_row in self.passenger_list:
                if passenger_row["client_id"] != client_row["client_id"]:
                    continue

                client_passenger = ClientPassenger(
                    p_name=passenger_row["passenger_module"],
                    p_queue_module=passenger_row["queue_module"],
                    p_sync_frequency=passenger_row["sync_frequency"],
                    p_queue_life_span=passenger_row["queue_life_span"])

                for puller_row in self.puller_list:
                    if ClientDataset._module_belongs_to_passenger(puller_row, passenger_row):
                        client_passenger.puller_modules.append(puller_row["puller_module"])
                for processor_row in self.processor_list:
                    if ClientDataset._module_belongs_to_passenger(processor_row, passenger_row):
                        client_passenger.processor_modules.append(processor_row["processor_module"])
                for pusher_row in self.pusher_list:
                    if ClientDataset._module_belongs_to_passenger(pusher_row, passenger_row):
                        client_passenger.pusher_modules.append(pusher_row["pusher_module"])
                client.passengers.append(client_passenger)

            for user_row in self.user_list:
                credential = Credential(
                    username=user_row["username"],
                    password=user_row["password"],
                    token=user_row["token"])
                if "role" in user_row:
                    role = str_to_role(user_row["role"])
                else:
                    role = Role.UNDEFINED
                user = User(credential=credential, role=role)
                client.users.append(user)

            output.append(client)
        return output

    def dump_as_dict(self) -> dict:
        """ Returns a dict """
        output = {
            "client": self.client_list,
            "passenger": self.passenger_list,
            "puller": self.puller_list,
            "processor": self.processor_list,
            "pusher": self.pusher_list,
            "user": self.user_list
        }
        return output

    def dump_as_json_string(self) -> str:
        """ Returns a string """
        dict_output = self.dump_as_dict()
        return json.dumps(dict_output, indent=4, sort_keys=True)

    def read_for_all_clients(self):
        """ Fills the dataset for all clients """
        self.client_list = self._query_helper.select_all_no_where("client")
        self.passenger_list = self._query_helper.select_all_no_where("passenger", p_order_by="exe_order") # pylint: disable= C0301
        self.puller_list = self._query_helper.select_all_no_where("puller", p_order_by="exe_order")
        self.processor_list = self._query_helper.select_all_no_where("processor", p_order_by="exe_order") # pylint: disable= C0301
        self.pusher_list = self._query_helper.select_all_no_where("pusher", p_order_by="exe_order")
        self.user_list = self._query_helper.select_all_no_where("webuser")

    def read_for_client(self):
        """ Fills the dataset for a single client """
        self.client_list = self._query_helper.select_all("client")
        self.passenger_list = self._query_helper.select_all("passenger", p_order_fields=["exe_order"]) # pylint: disable= C0301
        self.puller_list = self._query_helper.select_all("puller", p_order_fields=["exe_order"])
        self.processor_list = self._query_helper.select_all("processor", p_order_fields=["exe_order"]) # pylint: disable= C0301
        self.pusher_list = self._query_helper.select_all("pusher", p_order_fields=["exe_order"])
        self.user_list = self._query_helper.select_all("webuser")

    @staticmethod
    def _module_belongs_to_passenger(p_module_row: dict, p_passenger_row: dict) -> bool:
        if p_module_row["client_id"] != p_passenger_row["client_id"]:
            return False
        if p_module_row["passenger_id"] != p_passenger_row["passenger_id"]:
            return False
        return True


class SqlDatabase(AbstractDatabase):
    """ Implementation class for SQL Server database
    Keys of p_arguments can be found in databus.database.sql_db.sql_database_arguments
    """

    _LOG_ID_SEPARATOR = " | "

    def __init__(self,
                 p_client_id: str,
                 p_log: Log,
                 p_passenger_factory: AbstractPassengerFactory,
                 p_arguments: dict):
        super().__init__(p_client_id, p_log, p_passenger_factory, p_arguments)
        self._query_helper = QueryHelper(p_arguments, p_client_id)
        if p_client_id is None:
            self.client = None
        else:
            self.client = self._get_client(p_client_id)

    @property
    def client_master_data(self) -> str:
        """ Returns the definition of all clients as a JSON string """
        client_list = self._query_helper.select_all_no_where("client", "client_id")
        output_dict = {"client": client_list}
        return json.dumps(output_dict, indent=4, sort_keys=True)

    @client_master_data.setter
    def client_master_data(self, p_definitions: str):
        """ Writes the customizing into the database
        See getter for sample JSON format
        """
        ActionDecider(self._query_helper).decide_and_execute(
            json.loads(p_definitions),
            json.loads(self.client_master_data))

    @property
    def customizing(self) -> str:
        """ Returns the client customizing as JSON """
        dataset = ClientDataset(self._query_helper)
        dataset.read_for_client()
        return dataset.dump_as_json_string()

    @property
    def customizing_dict(self) -> dict:
        """ Returns the client customizing as dict """
        dataset = ClientDataset(self._query_helper)
        dataset.read_for_client()
        return dataset.dump_as_dict()

    @customizing.setter
    def customizing(self, p_customizing: str):
        """ Writes the customizing into the database
        See getter for sample JSON format
        """
        ActionDecider(self._query_helper).decide_and_execute(
            json.loads(p_customizing),
            self.customizing_dict)

    def delete_old_logs(self, p_before: datetime):
        """ Deletes old logs from the database """
        self.log.append_text("Deleting logs before " + p_before.isoformat())
        where_cond = WhereBuilder.date_lt("created_on", p_before)
        log_heads = self._query_helper.select_all("log_head", p_where=where_cond)

        try:
            for log_head in log_heads:
                self._query_helper.delete("log_item", p_where="log_id = '" + log_head["log_id"] + "'") # pylint: disable= C0301
                self._query_helper.delete("log_head", p_where="log_id = '" + log_head["log_id"] + "'") # pylint: disable= C0301
            self._query_helper.commit()
        except Exception as error:
            self._query_helper.rollback()
            raise error

    def delete_passenger_queue(self, p_passengers: List[AbstractPassenger]):
        """ Deletes passenger from the database """
        self.log.append_text("Deleting passengers from queue")

        try:
            for passenger in p_passengers:
                cond = "queue_id = '" + str(passenger.internal_id) + "'"
                self._delete_from_queue(p_where=cond)
            self._query_helper.commit()
        except Exception as error:
            self._query_helper.rollback()
            raise error

    def ensure_schema_existence(self):
        """ Checks the schema for the client
        Creates / completes the schema if anything is missing
        """
        existing_client = self._query_helper.select_single("client")
        if len(existing_client) > 0:
            return
        self._insert_into_client(self.client_id, 1)
        self._query_helper.commit()

    def erase_passenger_queue(self):
        """ Deletes all passengers from the database """
        self.log.append_text("Erasing passenger queue from the database")
        try:
            self._delete_from_queue()
            self._query_helper.commit()
        except Exception as error:
            self._query_helper.rollback()
            raise error

    def get_clients(self) -> List[Client]:
        return self._select_from_client()

    def get_log_content(self, p_log_id: str) -> str:
        """ Returns the contents of the given log
        p_log_id is whatever you have returned in get_log_list.
        """
        output = ""
        log_id = p_log_id.split(SqlDatabase._LOG_ID_SEPARATOR)[1]
        cond = "log_id = '" + log_id + "'"
        log_entries = self._query_helper.select_all("log_item", cond)

        for log_entry in log_entries:
            if output != "":
                output += "\r\n"

            line = Log.build_entry_field_string(
                p_date=str(log_entry["message_on"]),
                p_source=log_entry["module"],
                p_type=SqlToDatabus.message_type(log_entry["message_type"]).name,
                p_message=log_entry["message"])
            output += line

        return output

    def get_log_list(self) -> List[LogListItem]:
        """ Returns log entries """
        output = []
        log_list = self._query_helper.select_all("log_head")

        worst_msg_types = self._query_helper.select_all("log_worst_message_type")

        for log_entry in log_list:
            output_item = LogListItem()
            output_item.log_id = str(log_entry["created_on"]) + SqlDatabase._LOG_ID_SEPARATOR + log_entry["log_id"] # pylint: disable= C0301

            for worst_msg_type in worst_msg_types:
                if worst_msg_type["log_id"] == output_item.log_id:
                    output_item.worst_message_type = worst_msg_type["worst_message_type"]
                    break

            output.append(output_item)
        return output

    def get_passenger_queue_entries(self, # pylint: disable=R0913
                                    p_passenger_module: str = None,
                                    p_processor_status: QueueStatus = None,
                                    p_pusher_status: QueueStatus = None,
                                    p_puller_notified: bool = None,
                                    p_pulled_before: datetime = None
                                    ) -> List[PassengerQueueStatus]:
        """ Reads the requested entries from the disk """
        self.log.append_text("Reading passenger queue entries")
        return self._select_from_queue(
            p_passenger_module=p_passenger_module,
            p_processor_status=p_processor_status,
            p_pusher_status=p_pusher_status,
            p_puller_notified=p_puller_notified,
            p_pulled_before=p_pulled_before
        )

    def get_passenger_queue_entry(self, # pylint: disable=R0913
                                  p_internal_id: str
                                 ) -> PassengerQueueStatus:

        """ Reads a single entry """
        self.log.append_text("Reading passenger queue entry " + p_internal_id)
        return self._select_from_queue(p_queue_id=p_internal_id)[0]

    def insert_client(self, p_client: Client):
        """ Inserts a new client """
        try:
            self._insert_into_client(p_client.id, p_client.log_life_span)

            exe_order = 0
            for passenger in p_client.passengers:
                insert = InsertBuilder(self._query_helper.args, p_client.id)
                insert.table = "passenger"
                insert.add_string("passenger_id", passenger.id)
                insert.add_string("passenger_module", passenger.module)
                insert.add_string("queue_module", passenger.queue_module)
                insert.add_int("sync_frequency", passenger.sync_frequency)
                insert.add_int("queue_life_span", passenger.queue_life_span)
                exe_order += 1
                insert.add_int("exe_order", exe_order)
                self._query_helper.execute_insert(insert)

            for user in p_client.users:
                insert = InsertBuilder(self._query_helper.args, p_client.id)
                insert.table = "webuser"
                insert.add_string("username", user.credential.username)
                insert.add_string("password", user.credential.password)
                insert.add_string("token", user.credential.token)
                self._query_helper.execute_insert(insert)

            self._query_helper.commit()
        except Exception as error:
            self._query_helper.rollback()
            raise error

    def insert_log(self, p_log: Log):
        """ Creates a new log file on the disk """
        self.log.append_text("Writing log to the database")
        insert = InsertBuilder(self._query_helper.args, self.client_id)
        insert.table = "log_head"
        insert.add_string("log_id", p_log.guid)
        insert.add_datetime("created_on", p_log.creation_datetime)
        self._query_helper.execute_insert(insert)
        item_no = 0

        try:
            for log_entry in p_log.entries:
                item_no += 1
                insert.table = "log_item"
                insert.add_string("log_id", p_log.guid)
                insert.add_int("item_no", item_no)
                insert.add_string("module", log_entry.source)
                insert.add_string("message_type", DatabusToSql.message_type(log_entry.type))
                insert.add_string("message", log_entry.message)
                self._query_helper.execute_insert(insert)

            self._query_helper.commit()
        except Exception as error:
            self._query_helper.rollback()
            raise error

    def insert_passenger_queue(self, p_passenger_status: PassengerQueueStatus):
        """ Writes new files to the database """
        self.log.append_text("Appending passenger " + p_passenger_status.passenger.id_text)
        try:
            insert = InsertBuilder(self._query_helper.args, self.client_id)
            insert.table = "queue"
            insert.add_string("queue_id", p_passenger_status.passenger.internal_id)
            insert.add_string("external_id", p_passenger_status.passenger.external_id)
            insert.add_string("source_system", p_passenger_status.passenger.source_system)
            insert.add_string("passenger_module", p_passenger_status.passenger.passenger_module)
            insert.add_string("puller_module", p_passenger_status.passenger.puller_module)
            insert.add_string("puller_notified", DatabusToSql.boolean(False))
            insert.add_string("pulled_on", DatabusToSql.date_time(p_passenger_status.passenger.pull_datetime)) # pylint: disable= C0301
            self._query_helper.execute_insert(insert)

            processor_exe_order = 0
            for processor in p_passenger_status.processor_statuses:
                processor_exe_order += 1
                insert.table = "queue_processor"
                insert.add_string("queue_id", p_passenger_status.passenger.internal_id)
                insert.add_string("processor_module", processor.processor_module)
                insert.add_string("status", DatabusToSql.queue_status(processor.status))
                insert.add_int("exe_order", processor_exe_order)
                self._query_helper.execute_insert(insert)

            pusher_exe_order = 0
            for pusher in p_passenger_status.pusher_statuses:
                pusher_exe_order += 1
                insert.table = "queue_pusher"
                insert.add_string("queue_id", p_passenger_status.passenger.internal_id)
                insert.add_string("pusher_module", pusher.pusher_module)
                insert.add_string("status", DatabusToSql.queue_status(pusher.status))
                insert.add_int("exe_order", pusher_exe_order)
                self._query_helper.execute_insert(insert)

            for attachment in p_passenger_status.passenger.attachments:
                table_path = self._query_helper.path_builder.get_table_path("insert_queue_attachment") # pylint: disable=C0301

                sql = "{call " + table_path + "(?, ?, ?, ?, ?, ?)}"
                values = (self.client_id,
                          p_passenger_status.passenger.internal_id,
                          attachment.name,
                          attachment.text_content,
                          attachment.binary_content,
                          DatabusToSql.attachment_format(attachment.format))
                self._query_helper.execute_stored_procedure(sql, values)

            self._insert_logs(p_passenger_status)
            self._query_helper.commit()
        except Exception as error:
            self._query_helper.rollback()
            raise error

    def update_queue_status(self, p_status: PassengerQueueStatus):
        """ Updates queue files on the database """
        self.log.append_text("Updating passenger " + p_status.passenger.id_text)

        try:
            where = WhereBuilder(p_client_id=self.client_id)
            where.add_and("queue_id = '" + p_status.passenger.internal_id + "'")

            update = UpdateBuilder(self._query_helper.args)
            update.table = "queue"
            update.add_int("puller_notified", DatabusToSql.boolean(p_status.puller_notified))
            update.where = where
            self._query_helper.execute_update(update)

            for processor_status in p_status.processor_statuses:
                update.table = "queue_processor"
                update.add_string("status", DatabusToSql.queue_status(processor_status.status))
                update.where = where
                self._query_helper.execute_update(update)

            for pusher_status in p_status.pusher_statuses:
                update.table = "queue_pusher"
                update.add_string("status", DatabusToSql.queue_status(pusher_status.status))
                update.where = where
                self._query_helper.execute_update(update)

            self._insert_logs(p_status)
            self._query_helper.commit()

        except Exception as error:
            self._query_helper.rollback()
            raise error

    def convert_log_guid_to_id(self, p_guid: UUID) -> str:
        """ UUID to id conversion """
        guid_as_str = str(p_guid)
        log_ids = self.get_log_list()
        for log in log_ids:
            if guid_as_str in log.log_id:
                return log.log_id
        return ""

    def update_user_credential(self, p_credential: Credential):
        """ Updates the credential of the given user """
        where = WhereBuilder(p_client_id=self.client_id)
        where.add_and("username = '" + p_credential.username + "'")

        update = UpdateBuilder(self._query_helper.args)
        update.table = "webuser"
        update.add_string("password", p_credential.password)
        update.add_string("token", p_credential.token)
        update.where = where
        self._query_helper.execute_update(update)

    def _insert_logs(self, p_passenger_status: PassengerQueueStatus):
        del_where = "queue_id = '" + p_passenger_status.passenger.internal_id + "'"
        self._query_helper.delete("queue_log", p_where=del_where)

        insert = InsertBuilder(self._query_helper.args, self.client_id)
        for log in p_passenger_status.passenger.log_guids:
            insert.table = "queue_log"
            insert.add_string("queue_id", p_passenger_status.passenger.internal_id)
            insert.add_string("log_id", str(log))
            self._query_helper.execute_insert(insert)

    def _delete_from_queue(self, p_where: str = ""):
        self._query_helper.delete("queue_attachment", p_where)
        self._query_helper.delete("queue_processor", p_where)
        self._query_helper.delete("queue_pusher", p_where)
        self._query_helper.delete("queue_log", p_where)
        self._query_helper.delete("queue", p_where)

    def _get_client(self, p_id: str) -> Client:
        return self._select_from_client(p_id=p_id)[0]

    def _select_from_client(self, p_id: str = None) -> List[Client]: # pylint: disable=R0914
        client_dataset = ClientDataset(self._query_helper)
        if p_id is None:
            client_dataset.read_for_all_clients()
        else:
            client_dataset.read_for_client()
        return client_dataset.dump_as_client_list()

    def _select_from_queue(self, # pylint: disable=R0912, R0913, R0914, R0915
                           p_passenger_module: str = None,
                           p_processor_status: QueueStatus = None,
                           p_pusher_status: QueueStatus = None,
                           p_puller_notified: bool = None,
                           p_pulled_before: datetime = None,
                           p_queue_id: str = None
                           ) -> List[PassengerQueueStatus]:
        output = []
        where = WhereBuilder(self.client_id)

        if p_queue_id is not None:
            where.add_and("queue_id = '" + p_queue_id + "'")
        if p_passenger_module is not None:
            where.add_and("passenger_module = '" + p_passenger_module + "'")
        if p_puller_notified is not None:
            where.add_and("puller_notified = " + str(DatabusToSql.boolean(p_puller_notified)))
        if p_pulled_before is not None:
            where.add_and_date_lt("pulled_on", p_pulled_before)

        where.set_order_by(["pulled_on"])
        queue_list = self._query_helper.select_all_where_builder("queue", where)

        for queue_row in queue_list:
            queue_row_is_eligible = True
            queue_where = "queue_id = '" + queue_row["queue_id"] + "'"

            processor_list = self._query_helper.select_all("queue_processor", queue_where, p_order_fields=["exe_order"]) # pylint: disable= C0301
            if p_processor_status is not None and len(processor_list) > 0:
                for processor_row in processor_list:
                    if SqlToDatabus.queue_status(processor_row["status"]) != p_processor_status:
                        queue_row_is_eligible = False
                        break
            if not queue_row_is_eligible:
                continue

            pusher_list = self._query_helper.select_all("queue_pusher", queue_where, p_order_fields=["exe_order"]) # pylint: disable= C0301
            if p_pusher_status is not None and len(pusher_list) > 0:
                for pusher_row in pusher_list:
                    if SqlToDatabus.queue_status(pusher_row["status"]) != p_pusher_status:
                        queue_row_is_eligible = False
                        break
            if not queue_row_is_eligible:
                continue

            passenger = self.passenger_factory.create_passenger(queue_row["passenger_module"]) # pylint: disable=C0301
            passenger.external_id = queue_row["external_id"]
            passenger.internal_id = queue_row["queue_id"]
            passenger.source_system = queue_row["source_system"]
            passenger.puller_module = queue_row["puller_module"]
            passenger.pull_datetime = SqlToDatabus.date_time(queue_row["pulled_on"])

            attachment_list = self._query_helper.select_all("queue_attachment", queue_where)
            for attachment_row in attachment_list:
                if attachment_row["bin_content"] is None:
                    bin_content = None
                else:
                    bin_content = attachment_row["bin_content"]

                if attachment_row["txt_content"] is None:
                    txt_content = None
                else:
                    txt_content = attachment_row["txt_content"]

                attachment = Attachment(
                    p_name=attachment_row["attachment_id"],
                    p_format=SqlToDatabus.attachment_format(attachment_row["file_format"]),
                    p_text_content=txt_content,
                    p_binary_content=bin_content)
                passenger.attachments.append(attachment)

            log_list = self._query_helper.select_all("queue_log", queue_where)
            for log in log_list:
                passenger.collect_log_guid(UUID(log["log_id"]))

            output_row = PassengerQueueStatus(
                p_passenger=passenger,
                p_puller_notified=SqlToDatabus.boolean(queue_row["puller_notified"])
            )

            for processor_row in processor_list:
                pqs = ProcessorQueueStatus(
                    processor_row["processor_module"],
                    SqlToDatabus.queue_status(processor_row["status"]))
                output_row.processor_statuses.append(pqs)

            for pusher_row in pusher_list:
                pqs = PusherQueueStatus(
                    pusher_row["pusher_module"],
                    SqlToDatabus.queue_status(pusher_row["status"]))
                output_row.pusher_statuses.append(pqs)

            output.append(output_row)

        return output

    def _insert_into_client(self, p_client_id: str, p_log_life_span: int):
        insert = InsertBuilder(self._query_helper.args, p_client_id)
        insert.table = "client"
        insert.add_string("log_life_span", p_log_life_span)
        self._query_helper.execute_insert(insert)



""" Client module """
import datetime
from enum import Enum
from typing import List
from databus.client.client_passenger import ClientPassenger
from databus.client.user import Credential, User


class ClientError(Exception):
    """ Client exception class """

    class ErrorCode(Enum):
        """ Client error code """
        client_not_found = 1
        parameter_missing = 2
        authentication_error = 3

    def __init__(self, p_error_code: ErrorCode, p_client_id: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_client_id is None:
            self.client_id = ""
        else:
            self.client_id = p_client_id

    @property
    def message(self) -> str:
        """ Client error message """
        if self.error_code == ClientError.ErrorCode.client_not_found:
            return "Client " + self.client_id + " not found"
        if self.error_code == ClientError.ErrorCode.parameter_missing:
            return "Parameter missing, can't find client"
        if self.error_code == ClientError.ErrorCode.authentication_error:
            return "Invalid username or password"
        return "Client error"


class ClientPassengerError(Exception):
    """ Client passenger exception """

    class ErrorCode(Enum):
        """ Client passenger error code """
        passenger_not_found = 1

    def __init__(self,
                 p_error_code: ErrorCode,
                 p_client_id: str = None,
                 p_passenger_name: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_client_id is None:
            self.client_id = ""
        else:
            self.client_id = p_client_id

        if p_passenger_name is None:
            self.passenger_name = ""
        else:
            self.passenger_name = p_passenger_name

    @property
    def message(self) -> str:
        """ Client passenger error text """
        if self.error_code == ClientPassengerError.ErrorCode.passenger_not_found:
            return self.client_id + " doesn't contain passenger " + self.passenger_name
        return "Client passenger error: " + self.client_id + " " + self.passenger_name


class Client:
    """ Client class """

    ROOT = "root"

    def __init__(self,
                 p_id: str = "Undefined",
                 p_passengers: List[ClientPassenger] = None,
                 p_log_life_span: int = 0,
                 p_users: List[User] = None):
        self.id = p_id  # pylint: disable=C0103
        self.log_life_span = p_log_life_span

        if p_passengers is None:
            self.passengers = []
        else:
            self.passengers = p_passengers

        if p_users is None:
            self.users = []
        else:
            self.users = p_users

    @property
    def authorization_active(self) -> bool:
        """ Is authorization active or not? """
        return len(self.users) > 0

    @property
    def log_expiry_date(self) -> datetime.datetime:
        """ Log expiry date """
        return datetime.datetime.now() - datetime.timedelta(self.log_life_span)

    def authenticate(self, credential: Credential) -> User:
        """ Authenticates a user """
        if not self.authorization_active:
            return User()
        for user in self.users:
            if user.authenticate(credential):
                return user
        raise ClientPassengerError(
            ClientError.ErrorCode.authentication_error,
            self.id)

    def get_client_passenger(self, p_name: str) -> ClientPassenger:
        """ Returns the requested client passenger """
        for passenger in self.passengers:
            if passenger.name == p_name:
                return passenger

        raise ClientPassengerError(
            ClientPassengerError.ErrorCode.passenger_not_found,
            self.id,
            p_name)



""" Module for value conversions """
from abc import ABC, abstractstaticmethod
from databus.client.log import MessageType
from databus.passenger.attachment import AttachmentFormat
from databus.pqueue.queue_status import QueueStatus


class AbstractConverter(ABC):
    """ Abstract converter class """

    @abstractstaticmethod
    def attachment_format(input_val):
        """ Attachment format conversion """

    @abstractstaticmethod
    def boolean(input_val):
        """ Boolean value conversion """

    @abstractstaticmethod
    def date_time(input_val):
        """ Date conversion """

    @abstractstaticmethod
    def message_type(input_val):
        """ Message type conversion """

    @abstractstaticmethod
    def queue_status(input_val):
        """ Queue status conversion """


class DatabusToSql(AbstractConverter):
    """ Databus - SQL conversion """

    @staticmethod
    def attachment_format(input_val):
        """ Attachment format conversion """
        if input_val == AttachmentFormat.binary:
            return "B"
        if input_val == AttachmentFormat.text:
            return "T"
        raise Exception("Unknown attachment format " + str(input_val))

    @staticmethod
    def boolean(input_val):
        """ Boolean value conversion """
        if input_val:
            return 1
        return 0

    @staticmethod
    def date_time(input_val):
        """ Date conversion """
        year = str(input_val.year)
        month = DatabusToSql._get_numc(str(input_val.month), 2)
        day = DatabusToSql._get_numc(str(input_val.day), 2)
        hour = DatabusToSql._get_numc(str(input_val.hour), 2)
        minute = DatabusToSql._get_numc(str(input_val.minute), 2)
        second = DatabusToSql._get_numc(str(input_val.second), 2)
        return year + month + day + " " + hour + ":" + minute + ":" + second + ".000"

    @staticmethod
    def message_type(input_val):
        """ Message type conversion """
        if input_val == MessageType.error:
            return "E"
        if input_val == MessageType.info:
            return "I"
        if input_val == MessageType.undefined:
            return "U"
        if input_val == MessageType.warning:
            return "W"
        raise Exception("Unknown message type " + str(input_val))

    @staticmethod
    def queue_status(input_val):
        """ Queue status conversion """
        if input_val == QueueStatus.undefined:
            return "U"
        if input_val == QueueStatus.incomplete:
            return "I"
        if input_val == QueueStatus.complete:
            return "C"
        raise Exception("Unknown queue status " + str(input_val))

    @staticmethod
    def _get_numc(p_str: str, p_len: int) -> str:
        output = p_str
        while len(output) < p_len:
            output = "0" + output
        return output


class SqlToDatabus(AbstractConverter):
    """ SQL - Databus conversion """

    @staticmethod
    def attachment_format(input_val):
        """ Attachment format conversion """
        if input_val == "B":
            return AttachmentFormat.binary
        if input_val == "T":
            return AttachmentFormat.text
        raise Exception("Unknown attachment format " + input_val)

    @staticmethod
    def boolean(input_val):
        """ Boolean value conversion """
        return input_val == 1

    @staticmethod
    def date_time(input_val):
        """ Date conversion
        Current SQL driver returns native date time objects, so
        no conversion is needed.
        """
        return input_val

    @staticmethod
    def message_type(input_val):
        """ Message type conversion """
        if input_val == "E":
            return MessageType.error
        if input_val == "I":
            return MessageType.info
        if input_val == "U":
            return MessageType.undefined
        if input_val == "W":
            return MessageType.warning
        raise Exception("Unknown message type " + input_val)

    @staticmethod
    def queue_status(input_val):
        """ Queue status conversion """
        if input_val == "U":
            return QueueStatus.undefined
        if input_val == "I":
            return QueueStatus.incomplete
        if input_val == "C":
            return QueueStatus.complete
        raise Exception("Unknown queue status " + str(input_val))



""" Log module """
from datetime import datetime
from enum import Enum
import inspect
from typing import List
from uuid import uuid1, UUID


class MessageType(Enum):
    """ Message type """
    undefined = 0
    info = 1
    warning = 2
    error = 3


class LogEntry:
    """ Log entry class """
    def __init__(self,
                 p_message: str = None,
                 p_timestamp: datetime = datetime.now(),
                 p_type: MessageType = MessageType.info,
                 p_source: str = None):
        self._message = p_message
        self._timestamp = p_timestamp
        self._type = p_type

        if p_message is None:
            self._message = ""
        else:
            self._message = p_message

        if p_source == "" or p_source is None:
            frm = inspect.stack()[1]
            mod = inspect.getmodule(frm[0])
            self.source = mod.__name__
        else:
            self.source = p_source

    @property
    def message(self) -> str:
        """ Log entry as string """
        return self._message

    @property
    def timestamp(self) -> datetime:
        """ Log entry creation time """
        return self._timestamp

    @property
    def type(self) -> MessageType:
        """ Log entry message type """
        return self._type


class Log:
    """ Log class """
    def __init__(self):
        self._creation_datetime = datetime.now()
        self._guid = uuid1()
        self._entries = []

    @staticmethod
    def build_entry_field_string(p_date: str, p_source: str, p_type: str, p_message: str) -> str:
        """ Builds a string from entry fields """
        new_line = "[" + p_date + "]"
        new_line += "[" + p_source + "]"
        new_line += "[" + p_type + "]"
        new_line += " " + p_message
        return new_line

    @property
    def creation_datetime(self) -> datetime:
        """ Log creation time """
        return self._creation_datetime

    @property
    def entries(self) -> List[LogEntry]:
        """ All log entries """
        return self._entries

    @property
    def entries_as_string(self) -> str:
        """ Converts all log entries into string format """
        output = ""
        for entry in self.entries:
            new_line = Log.build_entry_field_string(
                entry.timestamp.isoformat(),
                entry.source,
                str(entry.type.name),
                entry.message)
            if output != "":
                output += "\r\n"
            output += new_line
        return output

    @property
    def guid(self) -> UUID:
        """ Unique log ID """
        return self._guid

    @property
    def has_error(self) -> bool:
        """ Returns true if there is any error message present """
        for entry in self._entries:
            if entry.type == MessageType.error:
                return True
        return False

    def append_entry(self, p_entry: LogEntry):
        """ Adds new entry to log """
        if p_entry.source == "" or p_entry.source is None:
            frm = inspect.stack()[1]
            mod = inspect.getmodule(frm[0])
            p_entry.source = mod.__name__
        self._entries.append(p_entry)

    def append_text(self, p_entry: str, p_source: str = None):
        """ Adds simple text to log """
        if p_source == "" or p_source is None:
            frm = inspect.stack()[1]
            mod = inspect.getmodule(frm[0])
            source = mod.__name__
        else:
            source = p_source
        self._entries.append(LogEntry(p_message=p_entry, p_source=source))



""" Module to build where conditions """
from datetime import datetime
from typing import List
from databus.database.sql_db.value_conversion import DatabusToSql


class WhereFlags: # pylint: disable=R0903
    """ Flags for where building """
    def __init__(self):
        self.has_where = False
        self.has_condition = False
        self.has_order_by = False

class WhereBuilder:
    """ Helper class to build where conditions """
    def __init__(self, p_client_id: str = ""):
        self._client_id = p_client_id
        self._flags = WhereFlags()
        self._where = ""
        self.clear()

    @staticmethod
    def date_lt(p_field: str, p_date: datetime):
        """ p_field < p_date """
        return p_field + " < '" + DatabusToSql.date_time(p_date) + "'"

    @staticmethod
    def int_eq(p_field: str, p_int: int):
        """ p_field = p_int """
        return p_field + " = " + str(p_int)

    @staticmethod
    def str_eq(p_field: str, p_str: str):
        """ p_field = 'p_str' """
        return p_field + " = '" + p_str + "'"

    @property
    def where(self) -> str:
        """ Where string """
        return self._where

    def add_and(self, p_condition: str):
        """ Adds a new and condition """
        if self._flags.has_order_by:
            raise Exception("Can't add conditions after ORDER BY")

        if p_condition is None or p_condition == "":
            return

        self._put_and()
        self._put_condition(p_condition)

    def add_and_date_lt(self, p_field: str, p_date: datetime):
        """ AND p_field > p_date """
        where_cond = WhereBuilder.date_lt(p_field, p_date)
        self.add_and(where_cond)

    def add_and_field_eq_int(self, p_field: str, p_int: int):
        """ AND p_field = p_int """
        where_cond = WhereBuilder.int_eq(p_field, p_int)
        self.add_and(where_cond)

    def add_and_field_eq_str(self, p_field: str, p_str: str):
        """ AND p_field = p_str """
        where_cond = WhereBuilder.str_eq(p_field, p_str)
        self.add_and(where_cond)

    def build(self, p_condition: str, p_order_fields: List[str] = None) -> str:
        """ Builds a new where condition """
        self.clear()
        self.add_and(p_condition)
        if p_order_fields is not None:
            self.set_order_by(p_order_fields)
        return self._where

    def build_without_client(self, p_condition: str) -> str:
        """ Builds a new where condition without using client id """
        self.clear(p_with_client=False)
        self._put_condition(p_condition)
        return self._where

    def clear(self, p_with_client: bool = True):
        """ Clears & starts anew """
        self._flags = WhereFlags()
        self._where = ""
        if p_with_client and self._client_id is not None and self._client_id != "":
            self._put_condition("client_id = '" + self._client_id + "'")

    def set_order_by(self, p_fields: List[str]):
        """ Order By """
        if self._flags.has_order_by:
            raise Exception("Can't set ORDER BY twice")

        if p_fields is None or len(p_fields) <= 0:
            return
        self._where += " ORDER BY "
        first_field = True
        for field in p_fields:
            if not first_field:
                self._where += ", "
            self._where += field
            first_field = False

        self._flags.has_order_by = True

    def _put_and(self):
        self._put_where()
        if self._flags.has_condition:
            self._where += " AND "

    def _put_condition(self, p_condition: str):
        self._put_where()
        self._where += " ( " + p_condition + " ) "
        self._flags.has_condition = True

    def _put_where(self):
        if self._flags.has_order_by:
            raise Exception("Can't put WHERE after ORDER BY")

        if not self._flags.has_where:
            self._where += " WHERE "
            self._flags.has_where = True



""" Abstract module to pull E-Mails from Exchange Server

If the expected E-Mails land to a single Exchange account, you can use
this module.

If the expected E-Mail can land on multiple Exchange accounts, you can use
databus.puller.abstract_multi_exchange .
"""
from abc import ABC, abstractmethod, abstractproperty
from datetime import datetime
from enum import Enum
from typing import List
from uuid import uuid1
from exchangelib import DELEGATE, Account, Configuration, Credentials, Mailbox, Message
from exchangelib.protocol import BaseProtocol, NoVerifyHTTPAdapter
from databus.client.log import Log, LogEntry, MessageType
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.attachment import Attachment, AttachmentFormat
from databus.passenger.email import Email
from databus.puller.abstract_puller import AbstractPuller


class ExchangeSettings: # pylint: disable=R0903
    """ Parameters to connect to Exchange Server """
    def __init__(self,
                 p_email: str = None,
                 p_username: str = None,
                 p_password: str = None,
                 p_server: str = None,
                 p_disable_ssl_validation: bool = None):
        if p_email is None:
            self.email = ""
        else:
            self.email = p_email

        if p_username is None:
            self.username = ""
        else:
            self.username = p_username

        if p_password is None:
            self.password = ""
        else:
            self.password = p_password

        if p_server is None:
            self.server = ""
        else:
            self.server = p_server

        if p_disable_ssl_validation is None:
            self.disable_ssl_validation = False
        else:
            self.disable_ssl_validation = p_disable_ssl_validation


class ExchangeFolderParent(Enum):
    """ Defines the possible parents of a folder """
    UNDEFINED = 0
    INBOX = 1
    ROOT = 2


class ExchangeFolder:
    """ Defines the path of a folder """
    def __init__(self,
                 p_parent: ExchangeFolderParent = ExchangeFolderParent.UNDEFINED,
                 p_path: str = ""):
        self.parent = p_parent
        self.path = p_path

    def get_folder_object(self, p_account: Account):
        """ Returns an Exchange folder object """
        # print(p_account.root.tree())
        if self.parent == ExchangeFolderParent.INBOX:
            return p_account.inbox / self.path
        if self.parent == ExchangeFolderParent.ROOT:
            return p_account.root / "Top of Information Store" / self.path
        raise Exception("Unexpected folder parent: " + self.parent.name)


class AbstractExchange(AbstractPuller, ABC):
    """ Abstract class to pull E-Mails from Exchange Server """
    _SOURCE_SYSTEM = "Exchange"
    _DEFAULT_EMAIL_MODULE = "databus.passenger.email"

    def __init__(self, p_log: Log = None, p_alias: str = ""):
        super().__init__(p_log)
        self._settings = self.settings
        self._email_decorator = None
        self.account = AbstractExchange._login(self._settings)
        self.email_module = AbstractExchange._DEFAULT_EMAIL_MODULE
        self.alias = p_alias

        self.log.append_text(
            "Exchange user: " +
            self._settings.username +
            " - " +
            self._settings.email)

    @abstractproperty
    def settings(self) -> ExchangeSettings:
        """ Returns parameters to connect to Exchange server """

    def delete_inbox_items(self, p_message_ids: List[str]):
        """ Deletes inbox items """
        for message_id in p_message_ids:
            self.log.append_text("Attempting to delete: " + message_id)
            found_in_inbox = False
            for inbox_item in self.account.inbox.all().order_by('-datetime_received'): # pylint: disable=E1101
                if message_id == inbox_item.message_id:
                    found_in_inbox = True
                    inbox_item.soft_delete()
                    self.log.append_text("Success!")
                    break
            if not found_in_inbox:
                self.log.append_entry(LogEntry(p_message="Item not found in inbox, assuming manual deletion", p_type=MessageType.warning)) # pylint: disable=C0301

    def forward_inbox_items(self, p_message_ids: List[str], p_recipients: List[str]):
        """ Forwards inbox items to the desired recipients """
        for message_id in p_message_ids:
            self.log.append_text("Attempting to forward: " + message_id)
            found_in_inbox = False
            for inbox_item in self.account.inbox.all().order_by('-datetime_received'): # pylint: disable=E1101
                if message_id == inbox_item.message_id:
                    found_in_inbox = True
                    inbox_item.forward(
                        subject="FWD: " + inbox_item.subject,
                        body="Forwarded by Databus",
                        to_recipients=p_recipients)
                    self.log.append_text("Success!")
                    break
            if not found_in_inbox:
                self.log.append_entry(LogEntry(p_message="Item not found in inbox, assuming manual deletion", p_type=MessageType.warning)) # pylint: disable=C0301

    def trash_inbox_items(self, p_message_ids: List[str]):
        """ Moves inbox items to trash """
        for message_id in p_message_ids:
            self.log.append_text("Attempting to trash: " + message_id)
            found_in_inbox = False
            for inbox_item in self.account.inbox.all().order_by('-datetime_received'): # pylint: disable=E1101
                if message_id == inbox_item.message_id:
                    found_in_inbox = True
                    inbox_item.move_to_trash()
                    self.log.append_text("Success!")
                    break
            if not found_in_inbox:
                self.log.append_entry(LogEntry(p_message="Item not found in inbox, assuming manual deletion", p_type=MessageType.warning)) # pylint: disable=C0301

    def move_inbox_items(self, p_message_ids: List[str], p_folder: ExchangeFolder):
        """ Moves inbox items to the desired folder """
        target_folder = p_folder.get_folder_object(self.account)

        for message_id in p_message_ids:
            self.log.append_text("Attempting to move: " + message_id)
            found_in_inbox = False
            for inbox_item in self.account.inbox.all().order_by('-datetime_received'): # pylint: disable=E1101
                if message_id == inbox_item.message_id:
                    found_in_inbox = True
                    inbox_item.move(target_folder)
                    self.log.append_text("Success!")
                    break
            if not found_in_inbox:
                self.log.append_entry(LogEntry(p_message="Item not found in inbox, assuming manual deletion", p_type=MessageType.warning)) # pylint: disable=C0301

    def delete_seated_passengers_from_inbox(self, p_seated_passengers: List[AbstractPassenger]):
        """ Deletes seated passengers from the inbox.
        This method is expected to be called from notify_passengers_seated in your concrete class.
        """
        message_ids = []
        for seated_passenger in p_seated_passengers:
            message_ids.append(seated_passenger.external_id)
        self.delete_inbox_items(message_ids)

    def forward_seated_passengers(self,
                                  p_seated_passengers: List[AbstractPassenger],
                                  p_recipients: List[str]):
        """ Forwards seated passengers to the desired recipients
        This method is expected to be called from notify_passengers_seated in your concrete class.
        """
        message_ids = []
        for seated_passenger in p_seated_passengers:
            message_ids.append(seated_passenger.external_id)
        self.forward_inbox_items(message_ids, p_recipients)
        self.delete_inbox_items(message_ids)

    def trash_seated_passengers(self, p_seated_passengers: List[AbstractPassenger]):
        """ Puts seated passengers into the recycle bin
        This method is expected to be called from notify_passengers_seated in your concrete class.
        """
        message_ids = []
        for seated_passenger in p_seated_passengers:
            message_ids.append(seated_passenger.external_id)
        self.trash_inbox_items(message_ids)

    def move_seated_passengers(self,
                               p_seated_passengers: List[AbstractPassenger],
                               p_folder: ExchangeFolder):
        """ Moves seated passengers to the desired folder
        This method is expected to be called from notify_passengers_seated in your concrete class.
        """
        message_ids = []
        for seated_passenger in p_seated_passengers:
            message_ids.append(seated_passenger.external_id)
        self.move_inbox_items(message_ids, p_folder)

    @abstractmethod
    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ This method is called whenever the E-Mail from Exchange is properly queued.
        When a passenger is seated, the E-Mail message shouldn't reside in the inbox.
        Otherwise, it would be processed again.
        You are advised to call one of the *_seated_passengers* methods from here.
        """

    def _pull(self) -> List[AbstractPassenger]:
        output = []

        for item in self.account.inbox.all().order_by('-datetime_received'):  # pylint: disable=E1101
            email_passenger = Email(p_external_id=item.message_id,
                                    p_internal_id=uuid1(),
                                    p_source_system=AbstractExchange._SOURCE_SYSTEM,
                                    p_attachments=[],
                                    p_puller_module=self.__module__,
                                    p_pull_datetime=datetime.now(),
                                    p_passenger_module=self.email_module)

            for item_attachment in item.attachments:
                try:
                    dummy = item_attachment.name
                    dummy = item_attachment.content
                except Exception: # pylint: disable=W0703
                    continue

                if any([item_attachment.name is None,
                        item_attachment.content is None]):
                    continue

                if any([item_attachment.content_type is None,
                        item_attachment.content_type == ""]):
                    attachment_format = Attachment.guess_format_by_file_name(
                        item_attachment.name)
                else:
                    attachment_format = Attachment.guess_format_by_mime_type(
                        item_attachment.content_type)

                if attachment_format == AttachmentFormat.text:
                    passenger_attachment = Attachment(
                        p_name=item_attachment.name,
                        p_format=AttachmentFormat.text,
                        p_text_content=str(item_attachment.content)[2:])
                else:
                    passenger_attachment = Attachment(
                        p_name=item_attachment.name,
                        p_format=AttachmentFormat.binary,
                        p_binary_content=item_attachment.content)

                email_passenger.attachments.append(passenger_attachment)

            if self._email_decorator is not None:
                self._email_decorator(item, email_passenger)
            self.log.append_text("Got mail from Exchange: " + email_passenger.id_text)
            output.append(email_passenger)

        return output

    def pull(self) -> List[AbstractPassenger]:
        """ Reads E-Mails from the given Exchange Server account """
        self._email_decorator = None
        return self._pull()

    def pull_with_email_decorator(self, p_decorator) -> List[AbstractPassenger]:
        """ Pulls E-Mails, calling p_decorator() to modify
        received email objects. p_decorator takes two parameters:
        1) Exchange item
        2) databus.passenger.email.Email
        """
        self._email_decorator = p_decorator
        output = self._pull()
        self._email_decorator = None
        return output

    def send_email(self, to: [str], subject: str, body: str): # pylint: disable=C0103
        """ Sends an E-Mail via Exchange Server """
        to_recipients = []
        for recipient in to:
            to_recipients.append(Mailbox(email_address=recipient))

        email_message = Message(account=self.account,
                                folder=self.account.sent,
                                subject=subject,
                                body=body,
                                to_recipients=to_recipients)

        email_message.send_and_save()

    @staticmethod
    def _login(settings: ExchangeSettings) -> Account:
        if settings.disable_ssl_validation:
            BaseProtocol.HTTP_ADAPTER_CLS = NoVerifyHTTPAdapter

        credentials = Credentials(username=settings.username,
                                  password=settings.password)

        if settings.server == "":
            account = Account(primary_smtp_address=settings.email,
                              credentials=credentials,
                              autodiscover=True,
                              access_type=DELEGATE)
        else:
            config = Configuration(server=settings.server, credentials=credentials)

            account = Account(primary_smtp_address=settings.email,
                              credentials=credentials,
                              autodiscover=False,
                              config=config,
                              access_type=DELEGATE)

        return account
